/*
 * Copyright 1993-2010 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 */
//
// Source code in this file modified from Nvidia OpenCL tutorial: 'vector addition'
// (http://developer.download.nvidia.com/compute/cuda/3_0/sdk/website/OpenCL/website/samples.html)
//
#include "clDistanceTransform.h"

static const int iDEBUG_FILE = 0;

DistanceTransform_GPU::DistanceTransform_GPU()
{
  //Ptrs init
  ckKernel = 0;
  cSourceCL = 0;
  cpProgram = 0;
  cqCommandQueue = 0;
  cxGPUContext = 0;
  cmDevSrc = 0;
  cmDevDst = 0;

  // Start logs 
  shrSetLogFileName ("oclDistanceTransformLog.log");
  if(iDEBUG_FILE){shrLog("clDistanceTransform_GPU begin");}

  //Get an OpenCL platform
  cl_platform_id cpPlatform;
  cl_int ciErr1 = clGetPlatformIDs(1, &cpPlatform, NULL);

  if(iDEBUG_FILE){shrLog("clGetPlatformID...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clGetPlatformID, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  //Get the devices
  cl_device_id cdDevice;
  ciErr1 = clGetDeviceIDs(cpPlatform, CL_DEVICE_TYPE_GPU, 1, &cdDevice, NULL);
  if(iDEBUG_FILE){shrLog("clGetDeviceIDs...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clGetDeviceIDs, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  //Create the context
  cxGPUContext = clCreateContext(0, 1, &cdDevice, NULL, NULL, &ciErr1);
  if(iDEBUG_FILE){shrLog("clCreateContext...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clCreateContext, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  // Create a command-queue
  cqCommandQueue = clCreateCommandQueue(cxGPUContext, cdDevice, 0, &ciErr1);
  if(iDEBUG_FILE){shrLog("clCreateCommandQueue...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clCreateCommandQueue, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  // Read the OpenCL kernel in from source file
  const char* cSourceFile = "./OpenCL/clDistanceTransform.cl";
  if(iDEBUG_FILE){shrLog("oclLoadProgSource (%s)...\n", cSourceFile);} 
  size_t szKernelLength;
  char* cSourceCL = oclLoadProgSource(cSourceFile, "", &szKernelLength);

  // Create the program
  cpProgram = clCreateProgramWithSource(cxGPUContext, 1, (const char **)&cSourceCL, &szKernelLength, &ciErr1);
  if(iDEBUG_FILE){shrLog("clCreateProgramWithSource...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clCreateProgramWithSource, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  // Build the program with 'mad' Optimization option
  #ifdef MAC
      char* flags = "-cl-fast-relaxed-math -DMAC";
  #else
      char* flags = "-cl-fast-relaxed-math";
  #endif
  ciErr1 = clBuildProgram(cpProgram, 0, NULL, NULL, NULL, NULL);
  if(iDEBUG_FILE){shrLog("clBuildProgram...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clBuildProgram, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  // Create the kernel
  ckKernel = clCreateKernel(cpProgram, "DistanceTransform", &ciErr1);
  if(iDEBUG_FILE){shrLog("clCreateKernel (VectorAdd)...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clCreateKernel, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }
}

DistanceTransform_GPU::~DistanceTransform_GPU()
{
  if(iDEBUG_FILE){shrLog("Starting Cleanup...\n\n");}

  // Cleanup allocated objects
  if(cSourceCL)free(cSourceCL);
  if(ckKernel)clReleaseKernel(ckKernel);  
  if(cpProgram)clReleaseProgram(cpProgram);
  if(cqCommandQueue)clReleaseCommandQueue(cqCommandQueue);
  if(cxGPUContext)clReleaseContext(cxGPUContext);
  if(cmDevSrc)clReleaseMemObject(cmDevSrc);
  if(cmDevDst)clReleaseMemObject(cmDevDst);
  
  if(iDEBUG_FILE)
  {shrLogEx(LOGBOTH | CLOSELOG, 0, "GPU mode end\n");}
}

void DistanceTransform_GPU::doDistanceTransform(const Image<float>& imInput, Image<float>& imOutput)
{
  // Setup input data
  size_t dataLen = imInput.getDx()*imInput.getDy();
  void* src = (void*)imInput.getDataPtr();
  void* dst = (void*)imOutput.getDataPtr();

  // Set and log global and local work size dimensions
  size_t szLocalWorkSize = 256;
  // Round up to the nearest multiple of the LocalWorkSize
  size_t szGlobalWorkSize = shrRoundUp((int)szLocalWorkSize, (int)(dataLen));
  if(iDEBUG_FILE){
  shrLog("Global Work Size \t\t= %u\nLocal Work Size \t\t= %u\n# of Work Groups \t\t= %u\n\n", 
         szGlobalWorkSize, szLocalWorkSize, (szGlobalWorkSize % szLocalWorkSize + szGlobalWorkSize/szLocalWorkSize)); 
  }

  // Allocate the OpenCL buffer memory objects for source and result on the device GMEM
  cl_int ciErr1;
  cmDevSrc = clCreateBuffer(cxGPUContext, CL_MEM_READ_ONLY, sizeof(cl_float) * dataLen, NULL, &ciErr1);
  cl_int ciErr2;
  cmDevDst = clCreateBuffer(cxGPUContext, CL_MEM_WRITE_ONLY, sizeof(cl_float) * dataLen, NULL, &ciErr2);
  ciErr1 |= ciErr2;
  if(iDEBUG_FILE){shrLog("clCreateBuffer...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
    if(iDEBUG_FILE){shrLog("Error in clCreateBuffer, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
    throw std::exception("GPU FAIL");
  }

  // Set the Argument values
  int im_dx = (int)imInput.getDx();
  int im_dy = (int)imInput.getDy();
  ciErr1  = clSetKernelArg(ckKernel, 0, sizeof(cl_mem), (void*)&cmDevSrc);
  ciErr1 |= clSetKernelArg(ckKernel, 1, sizeof(cl_mem), (void*)&cmDevDst);
  ciErr1 |= clSetKernelArg(ckKernel, 2, sizeof(cl_int), (void*)&im_dx);
  ciErr1 |= clSetKernelArg(ckKernel, 3, sizeof(cl_int), (void*)&im_dy);
  if(iDEBUG_FILE){shrLog("clSetKernelArg 0 - 3...\n\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
      if(iDEBUG_FILE){shrLog("Error in clSetKernelArg, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
      throw std::exception("GPU FAIL");
  }

  // Asynchronous write of data to GPU device
  ciErr1 = clEnqueueWriteBuffer(cqCommandQueue, cmDevSrc, CL_FALSE, 0, sizeof(cl_float) * dataLen, src, 0, NULL, NULL);
  if(iDEBUG_FILE){shrLog("clEnqueueWriteBuffer (SrcA and SrcB)...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
      if(iDEBUG_FILE){shrLog("Error in clEnqueueWriteBuffer, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
      throw std::exception("GPU FAIL");
  }

  // Launch kernel
  ciErr1 = clEnqueueNDRangeKernel(cqCommandQueue, ckKernel, 1, NULL, &szGlobalWorkSize, &szLocalWorkSize, 0, NULL, NULL);
  if(iDEBUG_FILE){shrLog("clEnqueueNDRangeKernel (VectorAdd)...\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
      if(iDEBUG_FILE){shrLog("Error in clEnqueueNDRangeKernel, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
      throw std::exception("GPU FAIL");
  }

  // Synchronous/blocking read of results, and check accumulated errors
  ciErr1 = clEnqueueReadBuffer(cqCommandQueue, cmDevDst, CL_TRUE, 0, sizeof(cl_float) * dataLen, dst, 0, NULL, NULL);
  if(iDEBUG_FILE){shrLog("clEnqueueReadBuffer (Dst)...\n\n");} 
  if (ciErr1 != CL_SUCCESS)
  {
      if(iDEBUG_FILE){shrLog("Error in clEnqueueReadBuffer, Line %u in file %s !!!\n\n", __LINE__, __FILE__);}
      throw std::exception("GPU FAIL");
  }
}



