#include "Tests.h"

#include "Timer.h"
#include "OpenCL/clDistanceTransform.h"

#include <exception>
#include <vector>
#include <string>
#include <sstream>
#include <cassert>
#include <fstream>

template <class T>
class DistanceTransformSolver
{
public:
  virtual void solveDT(const Image<T>& imInput, Image<T>& imOutput) = 0;
};

template <class T>
class DistanceTransformSolver_CPU
  : public DistanceTransformSolver<T>
{
public:
  virtual void solveDT(const Image<T>& imInput, Image<T>& imOutput)
  {
    cppComputeDistanceTransformImage(imInput, imOutput);
  }
};

template <class T>
class DistanceTransformSolver_GPU
  : public DistanceTransformSolver<T>
{
public:
  virtual void solveDT(const Image<T>& imInput, Image<T>& imOutput)
  {
    m_DistanceTransform_GPU.doDistanceTransform(imInput, imOutput);
  }

private:
  DistanceTransform_GPU m_DistanceTransform_GPU;
};

template <class T>
class TimeStats
{
public:
  TimeStats(const std::string& dataName)
  {
    m_dataName = dataName;
  };

  void addTrialSize(size_t tS)
  {
    m_vSizeToTest.push_back(tS);
  }

  void doTestsWithDistanceTransform(DistanceTransformSolver<T>& dtSolver)
  {
    m_vTestResultSeconds.clear();
    for(size_t i = 0; i < m_vSizeToTest.size(); ++i)
    {
      size_t square_im_size = m_vSizeToTest[i];
      Image<float> im(square_im_size, square_im_size, 0.0f);
      Image<float> im_out(square_im_size, square_im_size);
      im.safeSet(0, 0, 1.0f);

      double dTimeBegin = getCurTime();
      dtSolver.solveDT(im, im_out);
      double dTimeEnd = getCurTime();

      m_vTestResultSeconds.push_back(dTimeEnd-dTimeBegin);
    }
  }

  std::string getReport()
  {
    std::stringstream ss;
    ss<<m_dataName<<":"<<std::endl;
    for(size_t i = 0; i < m_vSizeToTest.size(); ++i)
    {
      ss<<"Size: "<<m_vSizeToTest[i]<<", run time: "<<m_vTestResultSeconds[i]<<"s"<<std::endl;
    }
    return ss.str();
  }

  void toMatlabFileAsVectors(const std::string& structName)
  {
    assert( m_vSizeToTest.size() == m_vTestResultSeconds.size() );

    std::stringstream ss_TimeTaken;
    ss_TimeTaken<<structName<<".vTimeTaken = [";
    std::stringstream ss_ImageSize;
    ss_ImageSize<<structName<<".vImageSize = [";

    for(size_t i = 0; i < m_vSizeToTest.size(); ++i)
    {
      ss_TimeTaken<<m_vTestResultSeconds[i]<<" ";
      ss_ImageSize<<m_vSizeToTest[i]<<" ";
    }

    ss_TimeTaken<<"];";
    ss_ImageSize<<"];";

    std::stringstream ssAll;
    ssAll<<ss_TimeTaken.str();
    ssAll<<std::endl;
    ssAll<<ss_ImageSize.str();
    ssAll<<std::endl;

    std::stringstream ssFilename;
    ssFilename<<"Stats_"<<structName<<".m";

    std::ofstream file_out;
    file_out.open(ssFilename.str().c_str());
    file_out<<ssAll.str();
    file_out.close();
  }  

private:
  std::string          m_dataName;
  std::vector<size_t>  m_vSizeToTest;
  std::vector<double>  m_vTestResultSeconds;
};

void test_DTisSame_CPUGPU()
{
  DistanceTransformSolver_CPU<float> dtSolver_CPU = DistanceTransformSolver_CPU<float>();
  DistanceTransformSolver_GPU<float> dtSolver_GPU = DistanceTransformSolver_GPU<float>();
  
  std::cout<<"Check whether the CPU-GPU yield (almost) the same DT result."<<std::endl;
  for(size_t i = 5; i < 25; i = i + 5)
  {
    //Setup input data
    size_t square_im_size = i;
    Image<float> im(square_im_size, square_im_size, 0.0f);
    im.safeSet(0, 0, 1.0f);
    
    Image<float> im_out_CPU(square_im_size, square_im_size);
    Image<float> im_out_GPU(square_im_size, square_im_size);
    
    //Solve
    dtSolver_CPU.solveDT(im, im_out_CPU);
    dtSolver_GPU.solveDT(im, im_out_GPU);

    //Check that the answers are the same
    if(!Image<float>::isEqual(im_out_CPU, im_out_GPU)) 
    {
      std::cout<<ImageToStr(im_out_CPU);
      std::cout<<ImageToStr(im_out_GPU);
      throw std::exception("GPU solution != GPU solution");
    }
    std::cout<<"   DT of "<<square_im_size<<"x"<<square_im_size<<" image OK"<<std::endl;
  }
}

void test_DistanceTransform_TimeStats()
{
  TimeStats<float> timeStats_CPU("CPU");
  timeStats_CPU.addTrialSize(25);
  timeStats_CPU.addTrialSize(50);
  timeStats_CPU.addTrialSize(75);
  timeStats_CPU.addTrialSize(100);
  timeStats_CPU.addTrialSize(125);
  timeStats_CPU.addTrialSize(150);
  timeStats_CPU.addTrialSize(175);
  timeStats_CPU.addTrialSize(200);
  timeStats_CPU.addTrialSize(225);
  timeStats_CPU.addTrialSize(250);
  timeStats_CPU.addTrialSize(275);
  timeStats_CPU.addTrialSize(300);
  timeStats_CPU.addTrialSize(325);
  timeStats_CPU.addTrialSize(350);
  timeStats_CPU.addTrialSize(375);
  timeStats_CPU.addTrialSize(400);
  timeStats_CPU.addTrialSize(425);
  timeStats_CPU.addTrialSize(450);

  TimeStats<float> timeStats_GPU("GPU");
  timeStats_GPU.addTrialSize(25);
  timeStats_GPU.addTrialSize(50);
  timeStats_GPU.addTrialSize(75);
  timeStats_GPU.addTrialSize(100);
  timeStats_GPU.addTrialSize(125);
  timeStats_GPU.addTrialSize(150);
  timeStats_GPU.addTrialSize(175);
  timeStats_GPU.addTrialSize(200);
  timeStats_GPU.addTrialSize(225);
  timeStats_GPU.addTrialSize(250);
  timeStats_GPU.addTrialSize(275);
  timeStats_GPU.addTrialSize(300);
  timeStats_GPU.addTrialSize(325);
  timeStats_GPU.addTrialSize(350);
  timeStats_GPU.addTrialSize(375);
  timeStats_GPU.addTrialSize(400);
  timeStats_GPU.addTrialSize(425);
  timeStats_GPU.addTrialSize(450);

  std::cout<<"Test the speed of DT on CPU"<<std::endl;
  timeStats_CPU.doTestsWithDistanceTransform(DistanceTransformSolver_CPU<float>());
  std::cout<<timeStats_CPU.getReport()<<std::endl;

  std::cout<<"Test the speed of DT on GPU"<<std::endl;
  timeStats_GPU.doTestsWithDistanceTransform(DistanceTransformSolver_GPU<float>());
  std::cout<<timeStats_GPU.getReport()<<std::endl;

  //Write results to HD
  timeStats_CPU.toMatlabFileAsVectors("CPU");
  timeStats_GPU.toMatlabFileAsVectors("GPU");
}

void test_initGetIm()
{
  Image<float> im(10, 10, 0.0f);
  im.safeSet(5, 5, 10.0f);

  size_t dx = im.getDx();
  size_t dy = im.getDy();

  float* dataPtr = im.getDataPtr();
  for(size_t x=0; x < dx; ++x)
  {
    for(size_t y=0; y < dy; ++y)
    {
      float res_get = im.safeGet(x,y);
      float res_ptr = dataPtr[dy*y + x];

      if(res_get != res_ptr)
      {
        throw std::exception("Image get interface problem");
      }
    }
  }
}

bool floatIsEqual(float f1, float f2)
{
  static const float verySmallDifference = 10e-20f;
  if( std::fabs(f1-f2) > verySmallDifference )
  {
    return false;
  }
  else
  {
    return true;
  }
}

bool floatIsAlmostEqual(float f1, float f2)
{
  static const float verySmallDifference = 10e-5f;
  if( std::fabs(f1-f2) > verySmallDifference )
  {
    return false;
  }
  else
  {
    return true;
  }
}

void test_DistanceTransform()
{
  Image<float> im(10, 10, 0.0f);
  Image<float> im_out(10, 10);

  im.safeSet(5, 5, 1.0f);

  cppComputeDistanceTransformImage(im, im_out);

  //std::cout<<"Init:"<<std::endl<<ImageToStr(im);
  //std::cout<<"DT:"<<std::endl<<ImageToStr(im_out);

  //Check known values by hand
  bool failedTest = false;
  
  if( !floatIsEqual(im_out.safeGet(5, 5), 0.0f) )
  {
    failedTest = true;
  }
  if( !floatIsEqual(im_out.safeGet(0, 5), 5.0f) )
  {
    failedTest = true;
  }
  if( !floatIsEqual(im_out.safeGet(0,0), 7.0710678118654755f) )
  {
    failedTest = true;
  }
  
  if(failedTest)
  {
    throw std::exception("Failed distance transform test");
  }
}

void test_CodeCorrectness()
{
  test_initGetIm();
  test_DistanceTransform();
  test_DTisSame_CPUGPU();
}
